package com.jdbernard.wdiwtlt.db

import java.sql.Connection
import java.sql.PreparedStatement
import java.sql.ResultSet
import java.text.SimpleDateFormat
import java.util.regex.Pattern
import javax.sql.DataSource

import groovy.sql.Sql
import groovy.transform.CompileStatic
import groovy.transform.TailRecursive

import com.jdbernard.wdiwtlt.db.models.*

public class ORM {

    private DataSource dataSource
    private Sql sql

    private static UPPERCASE_PATTERN = Pattern.compile(/(.)(\p{javaUpperCase})/)

    public ORM(DataSource dataSource) {
        this.dataSource = dataSource
        this.sql = new Sql(dataSource) }

    public void shutdown() { dataSource.shutdown() }

    /// ### Common
    public def getById(int id, Class modelClass) {
        def query = new StringBuilder()
            .append("SELECT * FROM ")
            .append(pluralize(nameFromModel(modelClass.simpleName)))
            .append(" WHERE id = ?")
            .toString()

        return recordToModel(sql.firstRow(query, id), modelClass) }

    public def getByName(String name, Class modelClass) {
        def query = new StringBuilder()
            .append("SELECT * FROM ")
            .append(pluralize(nameFromModel(modelClass.simpleName)))
            .append(" WHERE name = ?")
            .toString()

        return recordToModel(sql.firstRow(query, name), modelClass) }

    public def getBy(List<String> columns, List<Object> values,
    Class modelClass) {
        def query = new StringBuilder()
            .append("SELECT * FROM ")
            .append(pluralize(nameFromModel(modelClass.simpleName)))
            .append(" WHERE ")
            .append(columns.collect { it + " = ?" }.join(' AND '))
            .toString()

        return sql.rows(query, values)
            .collect { recordToModel(it, modelClass) } }

    public def save(def model) {
        if (model.id > 0) return update(model)
        else return create(model) }

    public def update(def model) {
        def setClauses = []
        def params = []

        model.class.fields
            .findAll { it.name != 'id' }
            .each { field ->
                setClauses << '"' + nameFromModel(field.name) + '"= ?'
                params << field.get(model) }

        def query = new StringBuilder()
            .append("UPDATE ")
            .append(pluralize(nameFromModel(model.class.simpleName)))
            .append(" SET ")
            .append(setClauses.join(', '))
            .append(" WHERE id = ")
            .append(model.id)
            .toString()

        return sql.executeUpdate(query, params) }

    public def create(def model) {
        def columns = []
        def params = []

        model.class.fields
            .findAll { it.name != 'id' }
            .each { field ->
                //if (field.class.getAnnotation(Model)) // check to see if we
                                                        // have nested models
                columns << '"' + nameFromModel(field.name) + '"'
                params << field.get(model) }

        def query= new StringBuilder()
            .append("INSERT INTO ")
            .append(pluralize(nameFromModel(model.class.simpleName)))
            .append(" (")
            .append(columns.join(', '))
            .append(") VALUES (")
            .append((1..columns.size()).collect { '?' }.join(', '))
            .append(")").toString()

        model.id = sql.executeInsert(query, params)[0][0]
        return 1 }

    public def delete(def model) {
        sql.execute("DELETE FROM ${model.class.simpleName} WHERE id = ?", [model.id])
        return sql.updateCount }

    public def associate(String linkTable, int firstId, int secondId) {
        return sql.executeInsert(
            "INSERT INTO $linkTable VALUES (?, ?) ON CONFLICT DO NOTHING",
            [firstId, secondId]) }

    public def associate(def m1, def m2) {
        return associate(pluralize(nameFromModel(m1.class.simpleName)) +
                   "_" + pluralize(nameFromModel(m2.class.simpleName)),
            m1.id, m2.id) }

    /// ### Album-specific methods
    public Album getAlbumById(int id) { return getById(id, Album) }
    public Album getAlbumByName(String name) { return getByName(name, Album) }
    public Album getAlbumByNameAndArtistId(String name, int artistId) {
        def albums = sql.rows("""\
            SELECT al.*
            FROM albums al JOIN
                 artists_albums aa ON
                    al.id = aa.album_id AND
                    aa.artist_id = ?
            WHERE al.name = ?""",
            [artistId, name]).collect { recordToModel(it, Album) }
        return albums ? albums[0] : null }

    public Album getAlbumsByArtistId(int artistId) {
        return sql.rows("""\
            SELECT al.*
            FROM albums al JOIN
                 artists_albums aa ON
                    al.id = aa.album_id AND
                    aa.artist_id = ?""",
            [artistId]).collect { recordToModel(it, Album) } }

    public List<Album> removeEmptyAlbums() {
        throw new UnsupportedOperationException("Not yet implemented.");
    }

    /// ### Artist-specific methods
    public Artist getArtistById(int id) { return getById(id, Artist) }
    public Artist getArtistByName(String name) { return getByName(name, Artist) }
    public Artist getArtistsByAlbum(int albumId) {
        return sql.rows("""\
            SELECT ar.*
            FROM artists ar JOIN
                 artists_albums aa ON
                    ar.id = aa.artist_id AND
                    aa.album_id = ?""",
            [albumId]).collect { recordToModel(it, Artist) } }

    public List<Artist> removeEmptyArtists() {
        throw new UnsupportedOperationException("Not yet implemented.");
    }

    public def addAlbumArtist(int albumId, int artistId) {
        return associate("artists_albums", artistId, albumId) }

    /// ### Bookmark-specific methods
    public Bookmark getBookmarkById(int id) { return getById(id, Bookmark) }
    public Bookmark getBookmarkByName(String name) { return getByName(name, Bookmark) }

    /// ### Image-specific methods
    public Image getImageById(int id) { return getById(id, Image) }

    /// ### MediaFile-specific methods
    public MediaFile getMediaFileById(int id) { return getById(id, MediaFile) }
    public MediaFile getMediaFileByName(String name) { return getByName(name, MediaFile) }

    public def associateMediaFileWithAlbum(int mediaFileId, int albumId) {
        return associate("albums_media_files", albumId, mediaFileId) }

    public def associateMediaFileWithArtist(int mediaFileId, int artistId) {
        return associate("artists_media_files", artistId, mediaFileId) }

    /// ### Playlist-specific methods
    public Playlist getPlaylistById(int id) { return getById(id, Playlist) }
    public Playlist getPlaylistByName(String name) { return getByName(name, Playlist) }

    public List<Playlist> removeEmptyPlaylists() {
        throw new UnsupportedOperationException("Not yet implemented.");
    }

    /// ### Tag-specific methods
    public Tag getTagById(int id) { return getById(id, Tag) }
    public Tag getTagByName(String name) { return getByName(name, Tag) }

    /// ### Utility functions
    public static String nameToModel(String name) {
        def pts = name.toLowerCase().split('_')
        return pts.length == 1 ? pts[0] :
            pts[0] + pts[1..-1].collect { it.capitalize() }.join() }

    public static String nameFromModel(String name) {
        return UPPERCASE_PATTERN.matcher(name).
            replaceAll(/$1_$2/).toLowerCase() }

    public static String pluralize(String name) { return name + "s" }

    static def updateModel(def record, def model) {
        model.class.fields.each { field ->
            field.set(model, record[nameFromModel(field.name)]) }
        return model }

    static def recordToModel(def record, Class clazz) {
        if (record == null) return null

        def model = clazz.newInstance()

        model.class.fields.each { field ->
            field.set(model, record[nameFromModel(field.name)]) }

        return model }

    static def modelToRecord(def model) {
        if (model == null) return null

        def record = [:]

        model.class.fields.each { field ->
            record[nameFromModel(field.name)] = field.get(model) }

        return record }
}
